/* packet-log3gpp.c
 * Routines for dissecting 3GPP protocol logs.
 * Copyright 2008, Vincent Helfre
 *
 * $Id: packet-log3gpp.c $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <glib.h>

#include <string.h>
#include <ctype.h>
#include <epan/packet.h>
#include <epan/emem.h>
#include <epan/proto.h>
#include <epan/strutil.h>
#include <wiretap/log3gpp.h>


/* Protocol and registered fields. */
static int proto_log3gpp = -1;


static int hf_log3gpp_timestamp = -1;
static int hf_log3gpp_protocol = -1;
static int hf_log3gpp_direction = -1;
static int hf_log3gpp_unparsed_data = -1;
static int hf_log3gpp_tty = -1;
static int hf_log3gpp_tty_line = -1;
static int hf_log3gpp_dissected_length = -1;

/* Protocol subtree. */
static int ett_log3gpp = -1;
static int ett_log3gpp_tty = -1;

typedef enum packet_direction_t
{
    UPLINK,
    DOWNLINK
} packet_direction_t;

static const value_string direction_vals[] = {
    { 0,   "Sent" },
    { 1,   "Received" },
    { 0,   NULL },
};
typedef struct
{
  char * protocol_name;
  char * ul_dissector_name;
  char * dl_dissector_name;
  dissector_handle_t ul_dissector_handle;
  dissector_handle_t dl_dissector_handle;
} lookup_dissector_element_t;


/* Look up table for protocol name /dissector: should be in alphabetic order!!!
   the purpose is to match a protocol name with a dissector,
   and to store the dissector handle the first time to avoid looking it up every time.
   This table should contain all 3GPP specified protocols */
lookup_dissector_element_t dissector_lookup_table[] = {
  {"GAN.TCP","umatcp","umatcp",0,0},
  {"GAN.UDP","umaudp","umaudp",0,0},
  {"GTP","gtp","gtp",0,0},
  {"LLC","llcgprs","llcgprs",0,0},
  {"LTE-RRC.BCCH.BCH","lte-rrc.bcch.bch","lte-rrc.bcch.bch",0,0},
  {"LTE-RRC.BCCH.DL.SCH","lte-rrc.bcch.dl.sch","lte-rrc.bcch.dl.sch",0,0},
  {"LTE-RRC.DL.CCCH","lte-rrc.dl.ccch","lte-rrc.dl.ccch",0,0},
  {"LTE-RRC.DL.DCCH","lte-rrc.dl.dcch","lte-rrc.dl.dcch",0,0},
  {"LTE-RRC.PCCH","lte-rrc.pcch","lte-rrc.pcch",0,0},
  {"LTE-RRC.UL.CCCH","lte-rrc.ul.ccch","lte-rrc.ul.ccch",0,0},
  {"LTE-RRC.UL.DCCH","lte-rrc.ul.dcch","lte-rrc.ul.dcch",0,0},
  {"RR","gsm_a_dtap","gsm_a_dtap",0,0},
  {"RRC.BCCH.BCH","rrc.bcch.bch","rrc.bcch.bch",0,0},
  {"RRC.BCCH.FACH","rrc.bcch.fach","rrc.bcch.fach",0,0},
  {"RRC.CCCH","rrc.ul.ccch","rrc.dl.ccch",0,0},
  {"RRC.DCCH","rrc.ul.dcch","rrc.dl.dcch",0,0},
  {"RRC.MCCH","rrc.mcch","rrc.mcch",0,0},
  {"RRC.MSCH","rrc.msch","rrc.msch",0,0},
  {"RRC.PCCH","rrc.pcch","rrc.pcch",0,0},
  {"RRC.SHCCH","rrc.ul.shcch","rrc.dl.shcch",0,0},
  {"SNDCP","sndcp","sndcp",0,0},
  {"SNDCPXID","sndcpxid","sndcpxid",0,0}
};

void proto_register_log3gpp(void);

int dissector_element_compare(const void *protocol_name, const lookup_dissector_element_t *element)
{
  return strcmp((const char *)protocol_name, ((const lookup_dissector_element_t *) element)->protocol_name);
}

dissector_handle_t look_for_dissector(char *protocol_name, packet_direction_t direction)
{
  lookup_dissector_element_t *element_ptr;
  dissector_handle_t dissector_handle = 0;

  element_ptr = (lookup_dissector_element_t *)bsearch((void *)protocol_name,
                                                      (void *)dissector_lookup_table, 
                                                      sizeof(dissector_lookup_table)/sizeof(lookup_dissector_element_t),
                                                      sizeof(lookup_dissector_element_t),
                                                      dissector_element_compare);
  if (element_ptr != NULL) {
    if (direction == UPLINK)
    {
      if (element_ptr->ul_dissector_handle == 0)
      {
        element_ptr->ul_dissector_handle = find_dissector(element_ptr->ul_dissector_name);
      }
      dissector_handle = element_ptr->ul_dissector_handle;
    }
    else
    {
      if (element_ptr->dl_dissector_handle == 0)
      {
        element_ptr->dl_dissector_handle = find_dissector(element_ptr->dl_dissector_name);
      }
      dissector_handle = element_ptr->dl_dissector_handle;
    }

  }
  else {
  }

  return dissector_handle;
}

/*****************************************/
/* Main dissection function.             */
/*****************************************/
static void
dissect_log3gpp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    proto_tree  *prot3gpp_tree = NULL;
    proto_item  *ti = NULL;
    gint        offset = 0;
    gint        protocol_start;
    gint        protocol_length;
    gint        timestamp_start;
    gint        timestamp_length;
    guint8      direction;
    tvbuff_t    *next_tvb;
    dissector_handle_t protocol_handle = 0;
    int sub_dissector_result = 0;
    char        *protocol_name;
    gboolean is_hex_data;

    /* Clear Info */
    if (check_col(pinfo->cinfo, COL_INFO))
    {
        col_clear(pinfo->cinfo, COL_INFO);
    }

    /* Create root (protocol) tree. */
    if (tree)
    {
        ti = proto_tree_add_item(tree, proto_log3gpp, tvb, offset, -1, FALSE);
        prot3gpp_tree = proto_item_add_subtree(ti, ett_log3gpp);
    }

    /*********************************************************************/
    /* Note that these are the fields of the stub header as written out  */
    /* by the wiretap module                                             */

    /* Timestamp in file */
    timestamp_start = offset;
    timestamp_length = tvb_strsize(tvb, offset);
    if (prot3gpp_tree) {
        proto_tree_add_double_format_value(prot3gpp_tree, hf_log3gpp_timestamp, tvb,
                                           offset, timestamp_length,
                                           atof(tvb_format_text(tvb, offset, timestamp_length)),
                                           "%s", tvb_format_text(tvb, offset, timestamp_length-1));
    }
    offset += timestamp_length;


    /* protocol name */
    protocol_start = offset;
    protocol_length = tvb_strsize(tvb, offset);
    if (prot3gpp_tree) {
        proto_tree_add_item(prot3gpp_tree, hf_log3gpp_protocol, tvb,
                            offset, protocol_length, FALSE);
    }
    offset += protocol_length;

    /* Direction */
    direction = tvb_get_guint8(tvb, offset);
    if (prot3gpp_tree) {
        proto_tree_add_item(prot3gpp_tree, hf_log3gpp_direction, tvb,
                            offset, 1, FALSE);
    }
    offset++;

    if (prot3gpp_tree)
    {
        /* Set selection length of prot3gpp tree */
        proto_item_set_len(prot3gpp_tree, offset);
    }

    /* Add useful details to protocol tree label */
    protocol_name = (char*)tvb_get_ephemeral_string(tvb, protocol_start, protocol_length);
    /* Set Protocol */
    if (check_col(pinfo->cinfo, COL_PROTOCOL))
    {
      col_set_str(pinfo->cinfo, COL_PROTOCOL, protocol_name);
    }
    /* To know whether the data following is row byte stream or text data */
    is_hex_data = strcmp(protocol_name,"TXT");

    if (tree)
    {
        proto_item_append_text(ti, " t=%s   %c   prot=%s",
                               tvb_get_ephemeral_string(tvb, timestamp_start, timestamp_length),
                               (direction == 0) ? 'U' : 'D',
                               protocol_name);
    }

    if (is_hex_data)
    {
      /* look up for the right dissector handle */
      protocol_handle = look_for_dissector(protocol_name, direction);

      /***********************************************************************/
      /* Now hand off to the dissector of intended packet encapsulation type */
      
      /* Try appropriate dissector, if one has been selected */
      if (protocol_handle != 0)
      {
        /* Dissect the remainder of the frame using chosen protocol handle */
        next_tvb = tvb_new_subset(tvb, offset, -1, tvb_reported_length(tvb)-offset);
        sub_dissector_result = call_dissector_only(protocol_handle, next_tvb, pinfo, tree);
      }
    }

    if (protocol_handle == 0 || sub_dissector_result == 0)
    {
      /* Could get here because:
         - desired protocol is unavailable (probably disabled), OR
         - protocol rejected our data
         Show remaining bytes as unparsed data */
      proto_tree_add_item(prot3gpp_tree, hf_log3gpp_unparsed_data,
                          tvb, offset, -1, FALSE);
      
      if (check_col(pinfo->cinfo, COL_INFO))
      {
        if (!is_hex_data)
        {
          col_add_fstr(pinfo->cinfo, COL_INFO,
                       "%s",
                       tvb_get_ephemeral_string(tvb, offset,tvb_reported_length(tvb)-offset));
        }
        else
        {
          col_add_fstr(pinfo->cinfo, COL_INFO,
                       "Not dissected  ( t=%s   %c   prot=%s)",
                       tvb_get_ephemeral_string(tvb, timestamp_start, timestamp_length),
                       (direction == 0) ? 'U' : 'D',
                       tvb_get_ephemeral_string(tvb, protocol_start, protocol_length));
        }
      }
    }
    else
    {
      /* Show number of dissected bytes */
      if (prot3gpp_tree) {
        proto_item *ti_local = proto_tree_add_uint(prot3gpp_tree,
                                                   hf_log3gpp_dissected_length,
                                                   tvb, 0, 0, tvb_reported_length(tvb)-offset);
        PROTO_ITEM_SET_GENERATED(ti_local);
      }
    }
}



/******************************************************************************/
/* Associate this protocol with the Catapult DCT2000 file encapsulation type. */
/******************************************************************************/
void proto_reg_handoff_log3gpp(void)
{
    dissector_handle_t log3gpp_handle = find_dissector("prot3gpp");
    dissector_add("wtap_encap", WTAP_ENCAP_LOG_3GPP,
                  log3gpp_handle);
}

/****************************************/
/* Register the protocol                */
/****************************************/
void proto_register_log3gpp(void)
{
    static hf_register_info hf[] =
    {
        { &hf_log3gpp_timestamp,
            { "Timestamp",
              "log3gpp.timestamp", FT_DOUBLE, BASE_DEC, NULL, 0x0,
              "File timestamp", HFILL
            }
        },
        { &hf_log3gpp_protocol,
            { "3GPP protocol",
              "log3gpp.protocol", FT_STRING, BASE_NONE, NULL, 0x0,
              "Original 3GPP protocol name", HFILL
            }
        },
        { &hf_log3gpp_direction,
            { "Direction",
              "log3gpp.direction", FT_UINT8, BASE_DEC, VALS(direction_vals), 0x0,
              "Frame direction (Sent or Received)", HFILL
            }
        },
        { &hf_log3gpp_unparsed_data,
            { "Unparsed protocol data",
              "log3gpp.unparsed_data", FT_BYTES, BASE_NONE, NULL, 0x0,
              "Unparsed 3GPP protocol data", HFILL
            }
        },
        { &hf_log3gpp_dissected_length,
            { "Dissected length",
              "log3gpp.dissected-length", FT_UINT16, BASE_DEC, NULL, 0x0,
              "Number of bytes dissected by subdissector(s)", HFILL
            }
        },

        { &hf_log3gpp_tty,
            { "tty contents",
              "log3gpp.tty", FT_NONE, BASE_NONE, NULL, 0x0,
              "tty contents", HFILL
            }
        },
        { &hf_log3gpp_tty_line,
            { "tty line",
              "log3gpp.tty-line", FT_STRING, BASE_NONE, NULL, 0x0,
              "tty line", HFILL
            }
        },
    };

    static gint *ett[] =
    {
        &ett_log3gpp,
        &ett_log3gpp_tty
    };

    /* Register protocol. */
    proto_log3gpp = proto_register_protocol("3GPP log packet",
                                            "LOG3GPP",
                                            "log3gpp");
    proto_register_field_array(proto_log3gpp, hf, array_length(hf));
    proto_register_subtree_array(ett, array_length(ett));

    /* Allow dissector to find be found by name. */
    register_dissector("prot3gpp", dissect_log3gpp, proto_log3gpp);
}

