/* log3gpp.c
 * Routines encapsulating/dumping 3gpp protocol logs.
 * The purpose of this format is to be able to log the 3GPP protocol stack on a mobile phone.
 * Copyright 2008, Vincent Helfre
 * $Id:  log3gpp.c $
 *
 * Wiretap Library
 * Copyright (c) 1998 by Gilbert Ramirez <gram@alumni.rice.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "wtap-int.h"
#include "file_wrappers.h"
#include "buffer.h"

#include "log3gpp.h"

#define MAX_FIRST_LINE_LENGTH      200
#define MAX_TIMESTAMP_LINE_LENGTH  100
#define MAX_LINE_LENGTH            65536
#define MAX_TIMESTAMP_LEN          32
#define MAX_SECONDS_CHARS          16
#define MAX_SUBSECOND_DECIMALS     4
#define MAX_PROTOCOL_NAME          64

/* 's' or 'r' of a packet as read from .out file */
typedef enum packet_direction_t
{
    uplink,
    downlink
} packet_direction_t;

gint first_packet_offset;
gchar firstline[MAX_FIRST_LINE_LENGTH];
gchar secondline[MAX_TIMESTAMP_LINE_LENGTH];
gint firstline_length = 0;
gint secondline_length = 0;

/***********************************************************/
/* Transient data used for parsing                         */

/* Buffer to hold a single text line read from the file */
static gchar linebuff[MAX_LINE_LENGTH];

/* 'Magic number' at start of 3gpp log files. */
static const gchar log3gpp_magic[] = "3GPP protocols transcript";

/* Protocol name of the packet that the packet was captured at */
static gchar protocol_name[MAX_PROTOCOL_NAME+1];

/************************************************************/
/* Functions called from wiretap core                       */
static gboolean log3gpp_read(wtap *wth, int *err, gchar **err_info,
                             gint64 *data_offset);
static gboolean log3gpp_seek_read(wtap *wth, gint64 seek_off,
                                  union wtap_pseudo_header *pseudo_header,
                                  guchar *pd, int length,
                                  int *err, gchar **err_info);
static void log3gpp_close(wtap *wth);

static gboolean log3gpp_dump(wtap_dumper *wdh, const struct wtap_pkthdr *phdr,
                             const union wtap_pseudo_header *pseudo_header,
                             const guchar *pd, int *err);
static gboolean log3gpp_dump_close(wtap_dumper *wdh, int *err);


/************************************************************/
/* Private helper functions                                 */
static gboolean read_new_line(FILE_T fh, gint64 *offset, gint *length);
static gboolean parse_line(gint line_length, gint *seconds, gint *useconds,
                           long *data_offset,
                           gint *data_chars,
                           packet_direction_t *direction,
                           gboolean *is_text_data);
static int write_stub_header(guchar *frame_buffer, char *timestamp_string,
                             packet_direction_t direction);
static guchar hex_from_char(gchar c);
static gchar char_from_hex(guchar hex);

static gboolean get_file_time_stamp(time_t *secs, guint32 *usecs);


/********************************************/
/* Open file (for reading)                 */
/********************************************/
int log3gpp_open(wtap *wth, int *err, gchar **err_info _U_)
{
    gint64  offset = 0;
    time_t  timestamp;
    guint32 usecs;

    /* Clear errno before reading from the file */
    errno = 0;

    /********************************************************************/
    /* First line needs to contain at least as many characters as magic */

    read_new_line(wth->fh, &offset, &firstline_length);

    if (((size_t)firstline_length < strlen(log3gpp_magic)) ||
        firstline_length >= MAX_FIRST_LINE_LENGTH)
    {
        return 0;
    }

    /* This file is not for us if it doesn't match our signature */
    if (memcmp(log3gpp_magic, linebuff, strlen(log3gpp_magic)) != 0)
    {
        return 0;
    }

    /***********************************************************/
    /* Second line contains file timestamp                     */
    read_new_line(wth->fh, &offset, &secondline_length);
    first_packet_offset = firstline_length +  secondline_length;

    if ((secondline_length >= MAX_TIMESTAMP_LINE_LENGTH) ||
        (!get_file_time_stamp(&timestamp, &usecs)))
    {
        /* Give up if file time line wasn't valid */
        return 0;
    }

    /* Allocate struct and fill in timestamp */
    wth->capture.log3gpp = g_malloc(sizeof(log3gpp_t));
    wth->capture.log3gpp->start_secs = timestamp;
    wth->capture.log3gpp->start_usecs = usecs;

    /************************************************************/
    /* File is for us. Fill in details so packets can be read   */

    /* Set our file type */
    wth->file_type = WTAP_FILE_LOG_3GPP;

    /* Use our own encapsulation to send all packets to our stub dissector */
    wth->file_encap = WTAP_ENCAP_LOG_3GPP;

    /* Callbacks for reading operations */
    wth->subtype_read = log3gpp_read;
    wth->subtype_seek_read = log3gpp_seek_read;
    wth->subtype_close = log3gpp_close;

    /* Choose microseconds (have 4 decimal places...) */
    wth->tsprecision = WTAP_FILE_TSPREC_USEC;

    *err = errno;

    return 1;
}


/**************************************************/
/* Read packet function.                          */
/* Look for and read the next usable packet       */
/* - return TRUE and details if found             */
/**************************************************/
gboolean log3gpp_read(wtap *wth, int *err, gchar **err_info _U_,
                               gint64 *data_offset)
{
    gint64 offset = wth->data_offset;
    long dollar_offset;
    packet_direction_t direction;
    gboolean is_text_data;

    /* Search for a line containing a usable packet */
    while (1)
    {
        int line_length, seconds, useconds, data_chars;
        gint64 this_offset = offset;

        /* Are looking for first packet after 2nd line */
        if (wth->data_offset == 0)
        {
            this_offset += first_packet_offset +1+1;
        }

        /* Clear errno before reading from the file */
        errno = 0;

        /* Read a new line from file into linebuff */
        if (read_new_line(wth->fh, &offset, &line_length) == FALSE)
        {
            /* Get out if no more lines can be read */
            break;
        }

        /* Try to parse the line as a frame record */
        if (parse_line(line_length, &seconds, &useconds,
                       &dollar_offset,
                       &data_chars, 
                       &direction,
                       &is_text_data))
        {
            guchar *frame_buffer;
            int n;
            int stub_offset = 0;
            char timestamp_string[MAX_TIMESTAMP_LEN+1];
            gint64 *pkey = NULL;

            g_snprintf(timestamp_string, 32, "%d.%04d", seconds, useconds/100);

            /* All packets go to 3GPP protocol stub dissector */
            wth->phdr.pkt_encap = WTAP_ENCAP_LOG_3GPP;

            /* Set data_offset to the beginning of the line we're returning.
               This will be the seek_off parameter when this frame is re-read.
            */
            *data_offset = this_offset;

            /* This is the position in the file where the next _read() will be called from */
            wth->data_offset = this_offset + line_length + 1;

            /* Fill in timestamp (capture base + packet offset) */
            wth->phdr.ts.secs = wth->capture.log3gpp->start_secs + seconds;
            if ((wth->capture.log3gpp->start_usecs + useconds) >= 1000000)
            {
                wth->phdr.ts.secs++;
            }
            wth->phdr.ts.nsecs =
                ((wth->capture.log3gpp->start_usecs + useconds) % 1000000) *1000;

            if (!is_text_data)
            {
              /* Get buffer pointer ready */
              buffer_assure_space(wth->frame_buffer,
                                  strlen(timestamp_string)+1 + /* timestamp */
                                  strlen(protocol_name)+1 +    /* Protocol name */
                                  1 +                          /* direction */
                                  (data_chars/2));
              frame_buffer = buffer_start_ptr(wth->frame_buffer); 
              
              /*********************/
              /* Write stub header */
              stub_offset = write_stub_header(frame_buffer, timestamp_string,
                                              direction);
  
              /* Binary data length is half bytestring length + stub header */
              wth->phdr.len = data_chars/2 + stub_offset;
              wth->phdr.caplen = data_chars/2 + stub_offset;
              /********************************/
              /* Copy packet data into buffer */
              for (n=0; n <= data_chars; n+=2)
              {
                frame_buffer[stub_offset + n/2] = (hex_from_char(linebuff[dollar_offset+n]) << 4) |
                                                   hex_from_char(linebuff[dollar_offset+n+1]);
              }
              *err = errno = 0;
              return TRUE;
            }
            else
            {
              /* Get buffer pointer ready */
              buffer_assure_space(wth->frame_buffer,
                                  strlen(timestamp_string)+1 + /* timestamp */
                                  strlen(protocol_name)+1 +    /* Protocol name */
                                  1 +                          /* direction */
                                  data_chars);
              frame_buffer = buffer_start_ptr(wth->frame_buffer);
 
              /*********************/
              /* Write stub header */
              stub_offset = write_stub_header(frame_buffer, timestamp_string,
                                              direction);

              /* Binary data length is bytestring length + stub header */
              wth->phdr.len = data_chars + stub_offset;
              wth->phdr.caplen = data_chars + stub_offset;

              /* do not convert the ascii char */
              memcpy(&frame_buffer[stub_offset],&linebuff[dollar_offset],data_chars);
              frame_buffer[stub_offset+data_chars-1]= '\0';
              *err = errno = 0;
              return TRUE;
            }
        }
    }

    /* No packet details to return... */
    *err = errno;
    return FALSE;
}


/**************************************************/
/* Read & seek function.                          */
/**************************************************/
static gboolean
log3gpp_seek_read(wtap *wth, gint64 seek_off,
                    union wtap_pseudo_header *pseudo_header, guchar *pd,
                    int length, int *err, gchar **err_info)
{
    gint64 offset = wth->data_offset;
    long dollar_offset;
    packet_direction_t direction;
    int seconds, useconds, data_chars;
    gboolean is_text_data;

    /* Reset errno */
    *err = errno = 0;

    /* Seek to beginning of packet */
    if (file_seek(wth->random_fh, seek_off, SEEK_SET, err) == -1)
    {
        return FALSE;
    }

    /* Re-read whole line (this really should succeed) */
    if (read_new_line(wth->random_fh, &offset, &length) == FALSE)
    {
        return FALSE;
    }

    /* Try to parse this line again (should succeed as re-reading...) */
    if (parse_line(length, &seconds, &useconds,
                   &dollar_offset,
                   &data_chars, 
                   &direction,
                   &is_text_data))
    {
        int n;
        int stub_offset = 0;
        char timestamp_string[32];
        g_snprintf(timestamp_string, 32, "%d.%04d", seconds, useconds/100);

        /* Make sure all packets go to log3gpp dissector */
        wth->phdr.pkt_encap = WTAP_ENCAP_LOG_3GPP;


        /*********************/
        /* Write stub header */
        stub_offset = write_stub_header((guchar*)pd, timestamp_string,
                                        direction);

        if (!is_text_data)
        {
          /********************************/
          /* Copy packet data into buffer */
          for (n=0; n <= data_chars; n+=2)
          {
            pd[stub_offset + n/2] = (hex_from_char(linebuff[dollar_offset+n]) << 4) |
                                     hex_from_char(linebuff[dollar_offset+n+1]);
          }
          *err = errno = 0;
          return TRUE;
        }
        else
        {
          /* do not convert the ascii char */
          memcpy(&pd[stub_offset],&linebuff[dollar_offset],data_chars);
          pd[stub_offset+data_chars-1] = '\0';
          *err = errno = 0;
          return TRUE;
        }
    }

    /* If get here, must have failed */
    *err = errno;
    *err_info = g_strdup_printf("prot 3gpp: seek_read failed to read/parse "
                                "line at position %" G_GINT64_MODIFIER "d",
                                seek_off);
    return FALSE;
}


/***************************************************************************/
/* Free log3gpp-specific capture info from file that was open for reading  */
/***************************************************************************/
void log3gpp_close(wtap *wth)
{
    /* Also free this capture info */
    g_free(wth->capture.log3gpp);
}

/***************************/
/* Dump functions          */
/***************************/

/*****************************************************/
/* The file that we are writing to has been opened.  */
/* Set other dump callbacks.                         */
/*****************************************************/
gboolean log3gpp_dump_open(wtap_dumper *wdh, gboolean cant_seek _U_, int *err _U_)
{
    /* Fill in other dump callbacks */
    wdh->subtype_write = log3gpp_dump;
    wdh->subtype_close = log3gpp_dump_close;

    return TRUE;
}

/*****************************************/
/* Write a single packet out to the file */
/*****************************************/

static gboolean do_fwrite(const void *data, size_t size, size_t count, FILE *stream, int *err_p)
{
    size_t nwritten;

    nwritten = fwrite(data, size, count, stream);
    if (nwritten != count) {
        if (nwritten == 0 && ferror(stream))
        {
            *err_p = errno;
        }
        else
        {
            *err_p = WTAP_ERR_SHORT_WRITE;
        }

        return FALSE;
    }
    return TRUE;
}

gboolean log3gpp_dump(wtap_dumper *wdh, const struct wtap_pkthdr *phdr,
                      const union wtap_pseudo_header *pseudo_header,
                      const guchar *pd, int *err)
{
    gchar time_string[16];
    guint32 n = 0;

    /******************************************************/
    /* Look up the file_externals structure for this file */
    /* Find wtap external structure for this wtap */

    if (wdh->dump.log3gpp == NULL)
    {
        /* Allocate the log3gpp-specific dump structure */
        wdh->dump.log3gpp = g_malloc(sizeof(log3gpp_t));

        /* Write out saved first line */
        if (! do_fwrite(firstline, 1, firstline_length, wdh->fh, err))
        {
            return FALSE;
        }
        if (! do_fwrite("\n", 1, 1, wdh->fh, err))
        {
            return FALSE;
        }

        /* Also write out saved second line with timestamp corresponding to the
           opening time of the log.
        */
        if (! do_fwrite(secondline, 1, secondline_length, wdh->fh, err))
        {
            return FALSE;
        }
        if (! do_fwrite("\n", 1, 1, wdh->fh, err))
        {
            return FALSE;
        }

        /* Allocate the log3gpp-specific dump structure */
        wdh->dump.log3gpp = g_malloc(sizeof(log3gpp_dump_t));

        /* Copy time of beginning of file */
        wdh->dump.log3gpp->start_time.secs =
            pseudo_header->log3gpp.wth->capture.log3gpp->start_secs;
        wdh->dump.log3gpp->start_time.nsecs =
            (pseudo_header->log3gpp.wth->capture.log3gpp->start_usecs * 1000);

        /* Set flag do don't write header out again */
        wdh->dump.log3gpp->first_packet_written = TRUE;
    }

    /* Calculate time of this packet to write, relative to start of dump */
    if (phdr->ts.nsecs >= wdh->dump.log3gpp->start_time.nsecs)
    {
        g_snprintf(time_string, 16, "%ld.%04d",
                 (long)(phdr->ts.secs - wdh->dump.log3gpp->start_time.secs),
                 (phdr->ts.nsecs - wdh->dump.log3gpp->start_time.nsecs) / 100000);
    }
    else
    {
        g_snprintf(time_string, 16, "%ld.%04u",
                 (long)(phdr->ts.secs - wdh->dump.log3gpp->start_time.secs-1),
                 ((1000000000 + (phdr->ts.nsecs / 100000)) - (wdh->dump.log3gpp->start_time.nsecs / 100000)) % 10000);
    }

    /* Write out the calculated timestamp */
    if (! do_fwrite(time_string, 1, strlen(time_string), wdh->fh, err))
    {
        return FALSE;
    }

    /****************************************************************/
    /* Need to skip stub header at start of pd before we reach data */

    /* Timestamp */
    for (; pd[n] != '\0'; n++);
    n++;

    /* Protocol name */
    for (; pd[n] != '\0'; n++);
    n++;

    /* Direction */
    n += 1;

    /**************************************/
    /* Remainder is encapsulated protocol */
    if (! do_fwrite("$", 1, 1, wdh->fh, err))
    {
        return FALSE;
    }

    /* Each binary byte is written out as 2 hex string chars */ 
    for (; n < phdr->len; n++)
    {
        gchar c[2];
        c[0] = char_from_hex((guchar)(pd[n] >> 4));
        c[1] = char_from_hex((guchar)(pd[n] & 0x0f));

        /* Write both hex chars of byte together */
        if (! do_fwrite(c, 1, 2, wdh->fh, err))
        {
            return FALSE;
        }
    }

    /* End the line */
    if (! do_fwrite("\n", 1, 1, wdh->fh, err))
    {
        return FALSE;
    }

    return TRUE;
}


/******************************************************/
/* Close a file we've been writing to.                */
/******************************************************/
static gboolean log3gpp_dump_close(wtap_dumper *wdh _U_, int *err _U_)
{
    return TRUE;
}




/****************************/
/* Private helper functions */
/****************************/

/**********************************************************************/
/* Read a new line from the file, starting at offset.                 */
/* - writes data to static var linebuff                               */
/* - on return 'offset' will point to the next position to read from  */
/* - return TRUE if this read is successful                           */
/**********************************************************************/
gboolean read_new_line(FILE_T fh, gint64 *offset, gint *length)
{
    char *result;

    /* Read in a line */
    result = file_gets(linebuff, MAX_LINE_LENGTH, fh);
    if (result == NULL)
    {
        /* No characters found */
        return FALSE;
    }

    /* Set length and offset.. */
    *length = strlen(linebuff);
    *offset = *offset + *length;

    /* ...but don't want to include newline in line length */
    if (linebuff[*length-1] == '\n')
    {
        linebuff[*length-1] = '\0';
        *length = *length - 1;
    }

    return TRUE;
}


/**********************************************************************/
/* Parse a line from buffer, by identifying:                          */
/* - context, port and direction of packet                            */
/* - timestamp                                                        */
/* - data position and length                                         */
/* Return TRUE if this packet looks valid and can be displayed        */
/**********************************************************************/
gboolean parse_line(gint line_length, gint *seconds, gint *useconds,
                    long *data_offset, gint *data_chars,
                    packet_direction_t *direction,
                    gboolean *is_text_data)
{
    int  n = 0;
    int  variant_digits = 0;
    int  variant = 1;
    int  protocol_chars = 0;

    char seconds_buff[MAX_SECONDS_CHARS+1];
    int  seconds_chars;
    char subsecond_decimals_buff[MAX_SUBSECOND_DECIMALS+1];
    int  subsecond_decimals_chars;

    /*********************************************************************/
    /* Find and read the timestamp                                       */
    /*********************************************************************/
    /* Now scan to the next digit, which should be the start of the timestamp */
    for (; !isdigit((guchar)linebuff[n]) && (n < line_length); n++);
    if (n >= line_length)
    {
        return FALSE;
    }

    /* Seconds */
    for (seconds_chars = 0;
         (linebuff[n] != '.') &&
         (seconds_chars <= MAX_SECONDS_CHARS) &&
         (n < line_length);
         n++, seconds_chars++)
    {
        if (!isdigit((guchar)linebuff[n]))
        {
            /* Found a non-digit before decimal point. Fail */
            return FALSE;
        }
        seconds_buff[seconds_chars] = linebuff[n];
    }
    if (seconds_chars > MAX_SECONDS_CHARS || n >= line_length)
    {
        /* Didn't fit in buffer.  Fail rather than use truncated */
        return FALSE;
    }

    /* Convert found value into number */
    seconds_buff[seconds_chars] = '\0';
    *seconds = atoi(seconds_buff);

    /* The decimal point must follow the last of the seconds digits */
    if (linebuff[n] != '.')
    {
        return FALSE;
    }
    /* Skip it */
    n++;

    /* Subsecond decimal digits (expect 4-digit accuracy) */
    for (subsecond_decimals_chars = 0;
         (linebuff[n] != ' ') &&
         (subsecond_decimals_chars <= MAX_SUBSECOND_DECIMALS) &&
         (n < line_length);
         n++, subsecond_decimals_chars++)
    {
        if (!isdigit((guchar)linebuff[n]))
        {
            return FALSE;
        }
        subsecond_decimals_buff[subsecond_decimals_chars] = linebuff[n];
    }
    if (subsecond_decimals_chars > MAX_SUBSECOND_DECIMALS || n >= line_length)
    {
        /* More numbers than expected - give up */
        return FALSE;
    }
    /* Convert found value into microseconds */
    subsecond_decimals_buff[subsecond_decimals_chars] = '\0';
    *useconds = atoi(subsecond_decimals_buff) * 100;

    /* Space character must follow end of timestamp */
    if (linebuff[n] != ' ')
    {
        return FALSE;
    }
    n++;

    /*********************************************************************/
    /* Find and read protocol name                                       */
    /*********************************************************************/
    /* Read context name until find ' ' */
    for (protocol_chars = 0;
         (linebuff[n] != ' ') && (protocol_chars < MAX_PROTOCOL_NAME) && (n < line_length);
         n++, protocol_chars++)
    {
        if (!isalnum((guchar)linebuff[n]) && linebuff[n] != '_' && linebuff[n] != '.' && linebuff[n] != '-')
        {
            return FALSE;
        }
        protocol_name[protocol_chars] = linebuff[n];
    }
    if (protocol_chars == MAX_PROTOCOL_NAME || n >= line_length)
    {
        /* If doesn't fit, fail rather than truncate */
        return FALSE;
    }
    protocol_name[protocol_chars] = '\0';

    /* Space char must follow protocol name */
    if (linebuff[n] != ' ')
    {
        return FALSE;
    }
    /* Skip it */
    n++;

    /* Scan ahead to the next space */
    for (; (!isalnum((guchar)linebuff[n])) && (n < line_length); n++);
    if (n >= line_length)
    {
        return FALSE;
    }


    if (strcmp(protocol_name,"TXT") == 0)
    {
      *direction = uplink;
      *data_offset = n;
      *data_chars = line_length - n;
      *is_text_data = TRUE;
    }
    else
    {
      /* Next character gives direction of message (must be 'u' or 'd') */
      if (linebuff[n] == 'u')
      {
        *direction = uplink;
      }
      else if (linebuff[n] == 'd')
      {
        *direction = downlink;
      }
      else
      {
        return FALSE;
      }
      
      /* Now skip ahead to find start of data (marked by '$') */
      for (; (linebuff[n] != '$') && (n+1 < line_length); n++);
      if (n+1 >= line_length)
      {
        return FALSE;
      }
      /* Skip it */
      n++;
    
    /* Set offset to data start within line */
    *data_offset = n;
    
    /* Set number of chars that comprise the hex string protocol data */
    *data_chars = line_length - n;
    
    *is_text_data = FALSE;
    }
    return TRUE;
}

/*****************************************************************/
/* Write the stub info to the data buffer while reading a packet */
/*****************************************************************/
int write_stub_header(guchar *frame_buffer, char *timestamp_string,
                      packet_direction_t direction)
{
    int stub_offset = 0;

    /* Timestamp within file */
    g_strlcpy((char*)&frame_buffer[stub_offset], timestamp_string, MAX_TIMESTAMP_LEN+1);
    stub_offset += (strlen(timestamp_string) + 1);

    /* Protocol name */
    g_strlcpy((char*)&frame_buffer[stub_offset], protocol_name, MAX_PROTOCOL_NAME+1);
    stub_offset += (strlen(protocol_name) + 1);

    /* Direction */
    frame_buffer[stub_offset] = direction;
    stub_offset++;

    return stub_offset;
}


/********************************************************/
/* Return hex nibble equivalent of hex string character */
/********************************************************/
guchar hex_from_char(gchar c)
{
    if ((c >= '0') && (c <= '9'))
    {
        return c - '0';
    }

    if ((c >= 'a') && (c <= 'f'))
    {
        return 0x0a + (c - 'a');
    }

    /* Not a valid hex string character */
    return 0xff;
}


/********************************************************/
/* Return character corresponding to hex nibble value   */
/********************************************************/
gchar char_from_hex(guchar hex)
{
    static char hex_lookup[16] =
    { '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};

    if (hex > 15)
    {
        return '?';
    }

    return hex_lookup[hex];
}

/************************************************************************/
/* Parse year, month, day, hour, minute, seconds out of formatted line. */
/* Set secs and usecs as output                                         */
/* Return FALSE if no valid time can be read                            */
/************************************************************************/
gboolean get_file_time_stamp(time_t *secs, guint32 *usecs)
{
    int n;
    struct tm tm;
    #define MAX_MONTH_LETTERS 9
    char month[MAX_MONTH_LETTERS+1];

    int day, year, hour, minute, second;
    int scan_found;

    /* If line longer than expected, file is probably not correctly formatted */
    if (strlen(linebuff) > MAX_TIMESTAMP_LINE_LENGTH)
    {
        return FALSE;
    }

    /**************************************************************/
    /* First is month. Read until get a space following the month */
    for (n=0; (linebuff[n] != ' ') && (n < MAX_MONTH_LETTERS); n++)
    {
        month[n] = linebuff[n];
    }
    month[n] = '\0';

    if      (strcmp(month, "January"  ) == 0)  tm.tm_mon = 0;
    else if (strcmp(month, "February" ) == 0)  tm.tm_mon = 1;
    else if (strcmp(month, "March"    ) == 0)  tm.tm_mon = 2;
    else if (strcmp(month, "April"    ) == 0)  tm.tm_mon = 3;
    else if (strcmp(month, "May"      ) == 0)  tm.tm_mon = 4;
    else if (strcmp(month, "June"     ) == 0)  tm.tm_mon = 5;
    else if (strcmp(month, "July"     ) == 0)  tm.tm_mon = 6;
    else if (strcmp(month, "August"   ) == 0)  tm.tm_mon = 7;
    else if (strcmp(month, "September") == 0)  tm.tm_mon = 8;
    else if (strcmp(month, "October"  ) == 0)  tm.tm_mon = 9;
    else if (strcmp(month, "November" ) == 0)  tm.tm_mon = 10;
    else if (strcmp(month, "December" ) == 0)  tm.tm_mon = 11;
    else
    {
        /* Give up if not found a properly-formatted date */
        return FALSE;
    }
    /* Skip space char */
    n++;

    /********************************************************/
    /* Scan for remaining numerical fields                  */
    scan_found = sscanf(linebuff+n, "%d, %d     %d:%d:%d.%u",
                        &day, &year, &hour, &minute, &second, usecs);
    if (scan_found != 6)
    {
        /* Give up if not all found */
        return FALSE;
    }

    /******************************************************/
    /* Fill in remaining fields and return it in a time_t */
    tm.tm_year = year - 1900;
    tm.tm_mday = day;
    tm.tm_hour = hour;
    tm.tm_min = minute;
    tm.tm_sec = second;
    tm.tm_isdst = -1;    /* daylight saving time info not known */

    /* Get seconds from this time */
    *secs = mktime(&tm);

    /* Multiply 4 digits given to get micro-seconds */
    *usecs = *usecs * 100;

    return TRUE;
}


